# pylint: disable=no-member, undefined-variable
from itertools import product

import pytest

import torch
from swyft.network import OnlineNormalizationLayer


class TestNormalizationLayer:
    bss = [1, 128]
    shapes = [
        (1,),
        (5,),
        (10, 5, 2, 1),
        (
            1,
            1,
            1,
        ),
    ]
    means = [2, 139]
    stds = [1, 80]
    stables = [True, False]

    @pytest.mark.parametrize(
        "bs, shape, mean, std, stable", product(bss, shapes, means, stds, stables)
    )
    def test_online_normalization_layer_update(self, bs, shape, mean, std, stable):
        onl = OnlineNormalizationLayer(shape, stable=stable)
        onl.train()
        old_stats = onl.n.clone(), onl.mean.clone(), onl.var.clone(), onl.std.clone()

        data = torch.randn(bs, *shape) * std + mean
        _ = onl(data)

        new_stats = onl.n.clone(), onl._mean.clone(), onl.var.clone(), onl.std.clone()

        assert old_stats != new_stats

    @pytest.mark.parametrize(
        "bs, shape, mean, std, stable", product(bss, shapes, means, stds, stables)
    )
    def test_online_normalization_layer_mean(self, bs, shape, mean, std, stable):
        torch.manual_seed(0)

        onl = OnlineNormalizationLayer(shape, stable=stable)
        onl.train()

        data = torch.randn(bs, *shape) * std + mean
        _ = onl(data)

        assert torch.allclose(onl.mean, data.mean(0))

    @pytest.mark.parametrize(
        "bs, shape, mean, std, stable", product(bss, shapes, means, stds, stables)
    )
    def test_online_normalization_layer_std(self, bs, shape, mean, std, stable):
        torch.manual_seed(0)

        onl = OnlineNormalizationLayer(shape, stable=stable)
        onl.train()

        data = torch.randn(bs, *shape) * std + mean
        _ = onl(data)

        if torch.isnan(data.std(0)).all():
            replacement_std = torch.sqrt(torch.ones_like(onl.std) * onl.epsilon)
            assert torch.allclose(onl.std, replacement_std)
        else:
            assert torch.allclose(onl.std, data.std(0))


if __name__ == "__main__":
    pass



# pylint: disable=no-member
import pytest
from itertools import product
import tempfile

import torch
import numpy as np
from swyft.intensity import (
    construct_intervals,
    get_factor_mask_from_intervals,
    Mask1d,
    FactorMask,
    Intensity,
)


class TestFactorMask:
    resolution = 100000

    xmaxes = [2 * np.pi, 4 * np.pi, 6 * np.pi]
    areas = [np.pi, 2 * np.pi, 3 * np.pi]

    @pytest.mark.parametrize("xmax, area", zip(xmaxes, areas))
    def test_get_factor_mask_from_intervals_1dim(self, xmax, area):
        x = np.linspace(0, xmax, TestFactorMask.resolution)
        y = np.sin(x)
        intervals = construct_intervals(x, y)
        factor_mask = get_factor_mask_from_intervals(intervals)
        assert np.isclose(
            area, factor_mask.area(), rtol=1e-4
        ), f"Truth {area} != factor mask {factor_mask.area()}."

    fns = [
        np.sin,
        np.cos,
        lambda x: -((x - 0.5) ** 2) + 0.1,
    ]
    area_over_4pi = [
        2 * np.pi,
        2 * np.pi,
        0.816228 - 0.183772,
    ]
    fn_area = list(zip(fns, area_over_4pi))

    @pytest.mark.parametrize("fn_area1, fn_area2", product(fn_area, fn_area))
    def test_get_factor_mask_from_intervals_2dim(self, fn_area1, fn_area2):
        fn1, area1 = fn_area1
        fn2, area2 = fn_area2

        x = np.linspace(0, 4 * np.pi, TestFactorMask.resolution)
        y1 = fn1(x)
        y2 = fn2(x)
        area = area1 * area2

        intervals = [construct_intervals(x, y) for y in (y1, y2)]
        factor_mask = get_factor_mask_from_intervals(intervals)
        assert np.isclose(
            area, factor_mask.area(), rtol=1e-4
        ), f"Truth {area} != factor mask {factor_mask.area()}."


class TestConstructIntervals:
    xmaxs = [4 * np.pi, 4 * np.pi, 1.0, 1.0, 1.0]
    fns = [
        np.sin,
        np.cos,
        lambda x: (x - 0.5) ** 2 - 0.1,
        lambda x: -((x - 0.5) ** 2) + 0.1,
        lambda x: x - 0.5,
    ]
    targets = [
        [[0.0, 3.140650081536511], [6.282556925810732, 9.424463770084952]],
        [
            [1.5696966593994006, 4.711603503673621],
            [7.853510347947842, 10.995417192222062],
        ],
        [[0.18371837183718373, 0.8161816181618162]],
        [[0.18371837183718373, 0.8161816181618162]],
        [[0.4999499949995, 1.0]],
    ]

    @pytest.mark.parametrize("fn, target, xmax", zip(fns, targets, xmaxs))
    def test_construct_intervals(self, fn, target, xmax):
        x = np.linspace(0, xmax, 10000)
        y = fn(x)
        intervals = construct_intervals(x, y)
        intervals = np.asarray(intervals)
        target = np.asarray(target)
        assert np.allclose(target, intervals)


class TestIntensityIO:
    expected_ns = [0, 1, 100]
    intervals = [
        [[0.0, 3.140650081536511], [6.282556925810732, 9.424463770084952]],
        [[0.18371837183718373, 0.8161816181618162]],
    ]

    @pytest.mark.parametrize("expected_n, intervals", product(expected_ns, intervals))
    def test_intensity_save_load(self, expected_n, intervals):
        factor_mask = get_factor_mask_from_intervals(intervals)
        intensity = Intensity(expected_n, factor_mask)
        with tempfile.NamedTemporaryFile() as tf:
            intensity.save(tf.name)

            loaded = Intensity.load(tf.name)

        gather_to_check = lambda x: (x.expected_n, x.area, x.factor_mask.intervals)
        objs = list(map(gather_to_check, [loaded, intensity]))
        assert [np.isclose(i, j) for i, j in zip(objs[0], objs[1])]



from .cache import Cache, DirectoryCache, MemoryCache
from .estimation import RatioEstimator, Points
from .intensity import get_unit_intensity, get_constrained_intensity
from .network import OnlineNormalizationLayer
from .plot import cont2d, plot1d, corner
from .train import get_statistics
from .utils import set_device, get_2d_combinations, cred1d

__all__ = [
    "Cache",
    "DirectoryCache",
    "MemoryCache",
    "RatioEstimator",
    "Points",
    "get_unit_intensity",
    "get_constrained_intensity",
    "OnlineNormalizationLayer",
    "cont2d",
    "plot1d",
    "corner",
    "get_statistics",
    "set_device",
    "get_2d_combinations",
    "cred1d",
    "run",
]


def run(
    x0,
    simulator,
    zdim,
    noise=None,
    cache=None,
    n_train=10000,
    n_rounds=3,
    device="cpu",
    max_epochs=10,
    batch_size=16,
    lr_schedule=[1e-3, 1e-4],
    threshold=1e-5,
    early_stopping_patience=1,
):
    """Default training loop. Possible to call just from observation x0 and simulator. Optionally, can tweak training details."""
    if cache is None:
        cache = MemoryCache(zdim=zdim, xshape=x0.shape)
    intensities = []
    res = []
    for i in range(n_rounds):
        if i == 0:
            intensity = get_unit_intensity(expected_n=n_train, dim=zdim)
        else:
            intensity = get_constrained_intensity(
                expected_n=n_train,
                ratio_estimator=res[-1],
                x0=x0,
                threshold=threshold,
            )
        intensities.append(intensity)
        cache.grow(intensities[-1])
        cache.simulate(simulator)
        points = Points(cache, intensities[-1], noise)
        re = RatioEstimator(points, device=device)
        res.append(re)
        res[-1].train(
            max_epochs=max_epochs,
            batch_size=batch_size,
            lr_schedule=lr_schedule,
            early_stopping_patience=early_stopping_patience,
        )
    return points, res[-1]



# pylint: disable=no-member, not-callable, access-member-before-definition
import math

import torch
import torch.nn as nn


def combine(y, z):
    """Combines data vectors y and parameter vectors z.

    z : (..., pnum, pdim)
    y : (..., ydim)

    returns: (..., pnum, ydim + pdim)

    """
    y = y.unsqueeze(-2)  # (..., 1, ydim)
    y = y.expand(*z.shape[:-1], *y.shape[-1:])  # (..., pnum, ydim)
    return torch.cat([y, z], -1)


class OnlineNormalizationLayer(nn.Module):
    def __init__(self, shape, stable: bool = False, epsilon: float = 1e-10):
        """Accumulate mean and variance online using the "parallel algorithm" algorithm from [1].

        Args:
            shape (tuple): shape of mean, variance, and std array. do not include batch dimension!
            stable (bool): (optional) compute using the stable version of the algorithm [1]
            epsilon (float): (optional) added to the computation of the standard deviation for numerical stability.

        References:
            [1] https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
        """
        super().__init__()
        self.register_buffer("n", torch.tensor(0, dtype=torch.long))
        self.register_buffer("_mean", torch.zeros(shape))
        self.register_buffer("_M2", torch.zeros(shape))
        self.register_buffer("epsilon", torch.tensor(epsilon))
        self.shape = shape
        self.stable = stable

    def _parallel_algorithm(self, x):
        assert x.shape[1:] == self.shape
        na = self.n.clone()
        nb = x.shape[0]
        nab = na + nb

        xa = self._mean.clone()
        xb = x.mean(dim=0)
        delta = xb - xa
        if self.stable:
            xab = (na * xa + nb * xb) / nab
        else:
            xab = xa + delta * nb / nab

        m2a = self._M2.clone()
        m2b = (
            x.var(dim=(0,), unbiased=False) * nb
        )  # do not use bessel's correction then multiply by total number of items in batch.
        m2ab = m2a + m2b + delta ** 2 * na * nb / nab
        return nab, xab, m2ab

    def forward(self, x):
        if self.training:
            self.n, self._mean, self._M2 = self._parallel_algorithm(x)
        return (x - self.mean) / self.std

    @property
    def mean(self):
        return self._mean

    @property
    def var(self):
        if self.n > 1:
            return self._M2 / (self.n - 1)
        else:
            return torch.zeros_like(self._M2)

    @property
    def std(self):
        return torch.sqrt(self.var + self.epsilon)


# From: https://github.com/pytorch/pytorch/issues/36591
class LinearWithChannel(nn.Module):
    def __init__(self, input_size, output_size, channel_size):
        super(LinearWithChannel, self).__init__()

        # initialize weights
        self.w = torch.nn.Parameter(torch.zeros(channel_size, output_size, input_size))
        self.b = torch.nn.Parameter(torch.zeros(channel_size, output_size))

        # change weights to kaiming
        self.reset_parameters(self.w, self.b)

    def reset_parameters(self, weights, bias):
        torch.nn.init.kaiming_uniform_(weights, a=math.sqrt(3))
        fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(weights)
        bound = 1 / math.sqrt(fan_in)
        torch.nn.init.uniform_(bias, -bound, bound)

    def forward(self, x):
        x = x.unsqueeze(-1)
        return torch.matmul(self.w, x).squeeze(-1) + self.b


class DenseLegs(nn.Module):
    def __init__(self, ydim, pnum, pdim=1, p=0.0, NH=256):
        super().__init__()
        self.fc1 = LinearWithChannel(ydim + pdim, NH, pnum)
        self.fc2 = LinearWithChannel(NH, NH, pnum)
        self.fc3 = LinearWithChannel(NH, NH, pnum)
        self.fc4 = LinearWithChannel(NH, 1, pnum)
        self.drop = nn.Dropout(p=p)

        self.af = torch.relu

        # swish activation function for smooth posteriors
        self.af2 = lambda x: x * torch.sigmoid(x * 10.0)

    def forward(self, y, z):
        x = combine(y, z)
        x = self.af(self.fc1(x))
        x = self.drop(x)
        x = self.af(self.fc2(x))
        x = self.drop(x)
        x = self.af(self.fc3(x))
        x = self.fc4(x).squeeze(-1)
        return x


class Network(nn.Module):
    def __init__(self, ydim, pnum, pdim=1, head=None, p=0.0, datanorms=None):
        """Base network combining z-independent head and parallel tail.

        :param ydim: Number of data dimensions going into DenseLeg network
        :param pnum: Number of posteriors to estimate
        :param pdim: Dimensionality of posteriors
        :param head: Head network, z-independent
        :type head: `torch.nn.Module`, optional

        The forward method of the `head` network takes data `x` as input, and
        returns intermediate state `y`.
        """
        super().__init__()
        # TODO make this handle yshape rather than ydim
        # TODO remove pnum and pdim
        self.head = head
        self.legs = DenseLegs(ydim, pnum, pdim=pdim, p=p)

        # Set datascaling
        if datanorms is None:
            datanorms = [
                torch.tensor(0.0),
                torch.tensor(1.0),
                torch.tensor(0.5),
                torch.tensor(0.5),
            ]
        self._set_datanorms(*datanorms)

    def _set_datanorms(self, x_mean, x_std, z_mean, z_std):
        self.x_loc = torch.nn.Parameter(x_mean)
        self.x_scale = torch.nn.Parameter(x_std)
        self.z_loc = torch.nn.Parameter(z_mean)
        self.z_scale = torch.nn.Parameter(z_std)

    def forward(self, x, z):
        x = (x - self.x_loc) / self.x_scale
        z = (z - self.z_loc) / self.z_scale

        if self.head is not None:
            y = self.head(x)
        else:
            y = x  # Use 1-dim data vector as features

        out = self.legs(y, z)
        return out


if __name__ == "__main__":
    pass



# pylint: disable=no-member, not-callable, undefined-variable
from pathlib import Path
import pickle

import torch.nn as nn
import numpy as np

from .types import Shape, Union, Sequence, PathType, Array, Device, Optional
from .eval import eval_net
from .utils import array_to_tensor, depth


class Intensity:
    """Intensity function constructed from a d-dim FactorMask."""

    def __init__(self, expected_n: int, factor_mask):
        """Creates a expected_n intensity function over the FactorMask.

        Args:
            expected_n (int): expected target number of samples
            factor_mask (FactorMask): d-dim FactorMask
        """
        self.expected_n = expected_n
        self.factor_mask = factor_mask
        self._area = None

    def __call__(self, z):
        return self.factor_mask(z) / self.area * self.expected_n

    def __repr__(self):
        return f"{self.__class__.__name__}({self.expected_n}, {repr(self.factor_mask)})"

    @property
    def area(self):
        if self._area is None:
            self._area = self.factor_mask.area()
        return self._area

    def sample(self, n=None):
        if n is None:
            n = np.random.poisson(self.expected_n)
        return self.factor_mask.sample(n)

    @property
    def intervals(self):
        return self.factor_mask.intervals

    @property
    def zdim(self):
        return self.factor_mask.dim

    def save(self, path: PathType):
        path = Path(path)
        with path.open("wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def load(path: PathType):
        path = Path(path)
        with path.open("rb") as f:
            obj = pickle.load(f)
        return obj


def get_unit_intensity(expected_n: int, dim: int):
    """Creates a expected_n intensity function over the unit dim hypercube.

    Args:
        expected_n: expected target number of samples
        dim: dim of hypercube
    """
    factor_mask = get_unit_factor_mask(dim)
    return Intensity(expected_n, factor_mask)


def constrain_intensity(
    expected_n: int,
    x0: Array,
    zdim: int,
    net: nn.Module,
    threshold: float,
    factor_mask=None,
    batch_size: int = 64,
    n_samples: int = 10000,
    device: Device = None,
):
    """Creates a expected_n intensity function constrained to ratio estimates over the treshold.
    The constraints are generated by sampling within a region. It is possible to restrict that search region using factor_mask.

    Args:
        expected_n: expected target number of samples
        x0: true observation
        zdim
        net: trained ratio estimator
        threshold: threshold for constraint
        factor_mask (FactorMask): (optional) factorized region, in which to search for constraints (default: hypercube)
        batch_size: evaluation minibatch size
        n_samples: number of samples to produce inorder to find the constrained intervals
        device
    """
    if factor_mask is None:
        factor_mask = get_unit_factor_mask(zdim)

    z = factor_mask.sample(n_samples)
    ratios = eval_net(
        x0=x0,
        net=net,
        z=np.expand_dims(z, axis=-1),
        batch_size=batch_size,
        device=device,
    )

    intervals_list = []
    for i in range(zdim):
        ratios_max = ratios[:, i].max()
        intervals = construct_intervals(
            z[:, i], ratios[:, i] - ratios_max - np.log(threshold)
        )
        intervals_list.append(intervals)

    factor_mask = get_factor_mask_from_intervals(intervals_list)
    print("Constrained posterior area:", factor_mask.area())
    return Intensity(expected_n, factor_mask)


def get_constrained_intensity(
    expected_n: int,
    ratio_estimator,
    x0: Array,
    threshold: float,
    factor_mask=None,
    batch_size: int = 64,
    n_samples: int = 10000,
):
    """Creates a expected_n intensity function constrained to ratio estimates over the treshold.
    The constraints are generated by sampling within a region. It is possible to restrict that search region using factor_mask.
    This is just like constrain_intensity, except ratio_estimator is introspected to get good default behavior.

    Args:
        expected_n: expected target number of samples
        ratio_estimator: takes signature ratio_estimator(x, z) and returns a likelihood ratio
        x0: true observation
        threshold: threshold for constraint
        factor_mask (FactorMask): (optional) factorized region, in which to search for constraints (default: ratio_estimator's)
        batch_size: evaluation minibatch size
        n_samples: number of samples to produce inorder to find the constrained intervals
    """
    return constrain_intensity(
        expected_n=expected_n,
        x0=x0,
        zdim=ratio_estimator.zdim,
        net=ratio_estimator.net,
        threshold=threshold,
        factor_mask=ratio_estimator.points.intensity.factor_mask,
        batch_size=batch_size,
        n_samples=n_samples,
        device=ratio_estimator.device,
    )


def construct_intervals(x, y):
    r"""Let x be the base space and let y be a scalar field over x. Get intervals \in x where y \geq 0.

    Returns a list of lists organized by:
    [[upcrossing, downcrossing], [upcrossing, downcrossing], ...]
    where upcrossing means y crosses from below zero to above zero in that region,
    and downcrossing means y crosses from above zero to below zero in that region.
    """
    assert x.shape == y.shape
    max_index = len(x) - 1

    indices = np.argsort(x)
    x = x[indices]
    y = y[indices]
    m = np.where(y >= 0.0, 1.0, 0.0)
    m = m[1:] - m[:-1]
    upcrossings = np.argwhere(m == 1.0)[:, 0]
    downcrossings = np.argwhere(m == -1.0)[:, 0]

    # TODO a y which is entirely above or below zero will return the whole interval. This is bad.
    # No crossings --> return entire interval
    if len(upcrossings) == 0 and len(downcrossings) == 0:
        return [[x[0], x[-1]]]

    if len(upcrossings) - len(downcrossings) == 1:
        # One more upcrossing than downcrossing
        # --> Treat right end as downcrossing
        downcrossings = np.append(downcrossings, max_index)
    elif len(upcrossings) - len(downcrossings) == -1:
        # One more downcrossing than upcrossing
        # --> Treat left end as upcrossing
        upcrossings = np.append(0, upcrossings)
    elif len(upcrossings) == len(downcrossings):
        pass
    else:
        raise ValueError(
            "It should be impossible to have two more crossings of one type, than the other."
        )

    intervals = []
    for down, up in zip(downcrossings, upcrossings):
        if up > down:
            intervals.append([x[down], x[up]])
        elif up < down:
            intervals.append([x[up], x[down]])
        elif up < 0 or down < 0:
            raise ValueError(
                "Constructing intervals with negative indexes is not allowed."
            )
        else:
            raise ValueError("Cannot have an up and down crossing at the same index.")
    return intervals


class Mask1d:
    """A 1-dim multi-interval based mask class."""

    def __init__(self, intervals: Sequence[Sequence[float]]):
        """One dimensional mask which indicates whether or not points lie within the region.

        Args:
            intervals (list of sorted tuples of floats): (lower, upper) bounds on regions which yield one upon evaluation. intervals with very small length are ignored.
        """
        increasing = lambda x, y: x < y
        zero_measure = lambda x, y: np.isclose(x, y, rtol=1e-8, atol=1e-10)
        measurable_intervals = [[x, y] for x, y in intervals if not zero_measure(x, y)]
        for x, y in measurable_intervals:
            assert increasing(
                x, y
            ), "One of the intervals was decreasing. They need to be (lower bound, upper bound)."
        self.intervals = np.asarray(measurable_intervals)  # n x 2 matrix

    def __call__(self, z):
        """Returns 1. if inside interval, otherwise 0."""
        m = np.zeros_like(z)
        for z0, z1 in self.intervals:
            zs_lt_z1 = z <= z1
            zs_gt_z0 = z >= z0
            ones_at_lt_z1 = np.where(zs_lt_z1, 1.0, 0.0)
            m += np.where(zs_gt_z0, ones_at_lt_z1, 0.0)
        assert not any(m > 1.0), "Overlapping intervals."
        return m

    def __repr__(self):
        return f"{self.__class__.__name__}({self.intervals.tolist()})"

    def area(self):
        """Combined length of all intervals (AKAK 1-dim area)."""
        return (self.intervals[:, 1] - self.intervals[:, 0]).sum()

    def sample(self, n):
        p = self.intervals[:, 1] - self.intervals[:, 0]
        p /= p.sum()
        i = np.random.choice(len(p), size=n, replace=True, p=p)
        w = np.random.rand(n)
        z = self.intervals[i, 0] + w * (self.intervals[i, 1] - self.intervals[i, 0])
        return z


class FactorMask:
    """A d-dim factorized mask."""

    def __init__(self, masks: Sequence):
        self.masks = masks
        self.dim = len(masks)

    def __call__(self, z):
        m = [self.masks[i](z[:, i]) for i in range(self.dim)]
        m = np.array(m).prod(axis=0)
        return m

    def __repr__(self):
        masks_as_string = ", ".join([repr(mask) for mask in self.masks])
        return f"{self.__class__.__name__}([{masks_as_string}])"

    def area(self):
        m = [self.masks[i].area() for i in range(self.dim)]
        return np.array(m).prod()

    def sample(self, n):
        z = np.empty((n, self.dim))
        for i in range(self.dim):
            z[:, i] = self.masks[i].sample(n)
        return z

    @property
    def intervals(self):
        return [mask.intervals for mask in self.masks]


def get_unit_factor_mask(dim):
    """Returns FactorMask on the dim unit hypercube."""
    return FactorMask([Mask1d([[0.0, 1.0]]) for _ in range(dim)])


def get_factor_mask_from_intervals(
    intervals: Union[Sequence[Sequence[Sequence[int]]], Sequence[Sequence[int]]]
):
    """Returns FactorMask from intervals. Intervals must have a depth of 2 or 3 (like np.array([]).ndim)."""
    _depth = depth(intervals)
    assert _depth in [2, 3], f"{_depth} was not 2 or 3."
    if _depth == 2:
        return FactorMask([Mask1d(intervals)])
    elif _depth == 3:
        return FactorMask([Mask1d(set_of_intervals) for set_of_intervals in intervals])
    else:
        raise ValueError(
            "Intervals must be a list of intervals or a list of a list of intervals."
        )



# pylint: disable=no-member
from typing import (
    Any,
    Callable,
    Dict,
    List,
    NewType,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    TYPE_CHECKING,
)
from pathlib import Path
from collections.abc import Collection

import numpy as np
import torch

PathType = Union[str, Path]

Device = Union[torch.device, str]
Dataset = torch.utils.data.Dataset

Tensor = torch.Tensor
Array = Union[np.ndarray, torch.Tensor]

Combinations = Union[int, Sequence[int], Sequence[Sequence[int]]]

Shape = Union[torch.Size, Tuple[int, ...]]
DictInt = Union[int, Dict[str, int]]
DictShape = Union[Shape, Dict[str, Shape]]



# pylint: disable=no-member
import torch
import torch.nn as nn
from numpy import ndarray

from .utils import combine_z, get_z, array_to_tensor
from .types import Array, Combinations, Device, Optional


def eval_net(
    x0: Array, net: nn.Module, z: Array, batch_size: int, device: Device = None
) -> ndarray:
    """Evaluate estimated likelihood ratios with a trained network.

    Args:
        x0 (Array): true observation
        net (nn.Module): trained ratio estimation network
        z (Array): shape (nsamples, pnum, pdim)
        batch_size (int): evaluation minibatch size
        device

    Returns:
        estimated likelihood ratio (ndarray): shape (nsamples, pnum)
    """
    net_was_training = net.training
    net.eval()

    nsamples = len(z)
    z = array_to_tensor(z, device=device)
    x0 = array_to_tensor(x0, device=device).unsqueeze(0)

    out = []
    for i in range(nsamples // batch_size + 1):
        zbatch = z[i * batch_size : (i + 1) * batch_size]
        out += net(x0, zbatch).detach().cpu()

    if net_was_training:
        net.train()
    return torch.stack(out).numpy()


def get_ratios(
    x0: Array,
    net: nn.Module,
    points: torch.utils.data.Dataset,
    combinations: Optional[Combinations] = None,
    batch_size: int = 64,
    device: Device = None,
    max_n_points: int = 1000,
) -> ndarray:
    """From parameter z estimate the corresponding likelihood ratios.

    Args:
        x0 (Array): true observation
        net (nn.Module): likelihood ratio estimator
        points: yields parameters to train on
        combinations: parameter combinations
        device
        batch_size (int): evaluation minibatch size
        max_n_points: number of points to evalute ratios on
    """
    x0 = array_to_tensor(x0, device=device)
    z = get_z(points)[:max_n_points]
    z = torch.stack(z)
    z = torch.stack([combine_z(zs, combinations) for zs in z])
    ratios = eval_net(x0, net, z, batch_size, device=device)
    return z.cpu().numpy(), ratios


if __name__ == "__main__":
    pass



import pytest

from functools import partial
from itertools import product
import tempfile
from copy import deepcopy

import torch
import torch.nn as nn
import numpy as np

from swyft.estimation import Points, RatioEstimator
from swyft.cache import MemoryCache
from swyft.intensity import get_unit_intensity


def sim_repeat_noise(theta, num_copies):
    noise = np.random.randn(num_copies, *theta.shape)
    expanded_theta = np.expand_dims(theta, axis=0)
    return expanded_theta + noise


def setup_points():
    zdim = 10
    num_copies = 3
    xshape = (num_copies, zdim)
    expected_n = 100
    simulator = partial(sim_repeat_noise, num_copies=num_copies)

    cache = MemoryCache(zdim, xshape)
    intensity = get_unit_intensity(expected_n, zdim)
    cache.grow(intensity)
    cache.simulate(simulator)
    return cache, Points(cache, intensity)


class Head(nn.Module):
    def __init__(self, in_features, out_features):
        super().__init__()
        self.flatten = torch.nn.Flatten()
        self.layer = torch.nn.Linear(in_features, out_features)

    def forward(self, x):
        return self.layer(self.flatten(x))

    @staticmethod
    def featurize(xshape):
        in_features = np.product(xshape)
        out_features = 100
        return in_features, out_features


class TestPoints:
    def test_points_save_load(self):
        cache, points = setup_points()
        with tempfile.NamedTemporaryFile() as tf:
            points.save(tf.name)

            loaded = Points.load(cache, tf.name)

        gather_attrs = lambda x: (
            x.indices,
            x.xshape,
            x.zdim,
            x.intensity.expected_n,
            x.intensity.area,
            x.intensity.factor_mask.intervals,
        )
        assert [
            np.allclose(i, j)
            for i, j in zip(gather_attrs(loaded), gather_attrs(points))
        ]


class TestRatioEstimator:
    @pytest.mark.parametrize("head", (None, Head))
    def test_ratio_estimator_save_load(self, head):
        cache, points = setup_points()

        if head is None:
            head1 = None
            head2 = None
        else:
            in_features, out_features = Head.featurize(points.xshape)
            head1 = head(in_features, out_features)
            head2 = head(in_features, out_features)

        re = RatioEstimator(points, head=head1)
        with tempfile.NamedTemporaryFile() as tf:
            re.save(tf.name)

            loaded = RatioEstimator.load(cache, tf.name, head=head2)

        gather_attrs = lambda x: (
            x.combinations,
            *(v for _, v in x.net_state_dict.items()),
            *(v for _, v in x.ratio_cache.items()),
            x.points.indices,
            x.points.xshape,
            x.points.zdim,
            x.points.intensity.expected_n,
            x.points.intensity.area,
            x.points.intensity.factor_mask.intervals,
        )
        assert [
            np.allclose(i, j) for i, j in zip(gather_attrs(loaded), gather_attrs(re))
        ]


if __name__ == "__main__":
    pass



# pylint: disable=no-member, not-callable
from copy import deepcopy
from contextlib import suppress

import numpy as np
import torch

from .utils import combine_z
from .types import Sequence, Combinations, Dict, Union, Array


def loss_fn(network, xz, combinations: Combinations = None):
    """Evaluate binary-cross-entropy loss function. Mean over batch.

    Args:
        network (nn.Module): network taking minibatch of samples and returing ratio estimator.
        xz (dict): batch of samples to train on.
        combinations (list, optional): determines posteriors that are generated.
            examples:
                [[0,1], [3,4]]: p(z_0,z_1) and p(z_3,z_4) are generated
                    initialize network with zdim = 2, pdim = 2
                [[0,1,5,2]]: p(z_0,z_1,z_5,z_2) is generated
                    initialize network with zdim = 1, pdim = 4

    Returns:
        Tensor: training loss.
    """
    assert xz["x"].size(0) == xz["z"].size(0), "Number of x and z must be equal."
    assert (
        xz["x"].size(0) % 2 == 0
    ), "There must be an even number of samples in the batch for contrastive learning."
    n_batch = xz["x"].size(0)

    # Is it the removal of replacement that made it stop working?!

    # bring x into shape
    # (n_batch*2, data-shape)  - repeat twice each sample of x - there are n_batch samples
    # repetition pattern in first dimension is: [a, a, b, b, c, c, d, d, ...]
    x = xz["x"]
    x = torch.repeat_interleave(x, 2, dim=0)

    # bring z into shape
    # (n_batch*2, param-shape)  - repeat twice each sample of z - there are n_batch samples
    # repetition is alternating in first dimension: [a, b, a, b, c, d, c, d, ...]
    z = xz["z"]
    z = torch.stack([combine_z(zs, combinations) for zs in z])
    zdim = len(z[0])
    z = z.view(n_batch // 2, -1, *z.shape[-1:])
    z = torch.repeat_interleave(z, 2, dim=0)
    z = z.view(n_batch * 2, -1, *z.shape[-1:])

    # call network
    lnL = network(x, z)
    lnL = lnL.view(n_batch // 2, 4, zdim)

    # Evaluate cross-entropy loss
    # loss =
    # -ln( exp(lnL(x_a, z_a))/(1+exp(lnL(x_a, z_a))) )
    # -ln( exp(lnL(x_b, z_b))/(1+exp(lnL(x_b, z_b))) )
    # -ln( 1/(1+exp(lnL(x_a, z_b))) )
    # -ln( 1/(1+exp(lnL(x_b, z_a))) )
    loss = -torch.nn.functional.logsigmoid(lnL[:, 0])
    loss += -torch.nn.functional.logsigmoid(-lnL[:, 1])
    loss += -torch.nn.functional.logsigmoid(-lnL[:, 2])
    loss += -torch.nn.functional.logsigmoid(lnL[:, 3])
    loss = loss.sum() / (n_batch // 2)

    return loss


def split_length_by_percentage(length: int, percents: Sequence[float]) -> Sequence[int]:
    assert np.isclose(sum(percents), 1.0), f"{percents} does not sum to 1."
    lengths = [int(percent * length) for percent in percents]

    # Any extra from round off goes to the first split.
    difference = length - sum(lengths)
    lengths[0] += difference
    assert length == sum(
        lengths
    ), f"Splitting into {lengths} should equal total {length}."
    return lengths


# We have the posterior exactly because our proir is known and flat. Flip bayes theorem, we have the likelihood ratio.
# Consider that the variance of the loss from different legs causes some losses to have high coefficients in front of them.
def train(
    network,
    train_loader,
    validation_loader,
    early_stopping_patience,
    max_epochs=None,
    lr=1e-3,
    combinations=None,
    device=None,
    non_blocking=True,
):
    """Network training loop.

    Args:
        network (nn.Module): network for ratio estimation.
        train_loader (DataLoader): DataLoader of samples.
        validation_loader (DataLoader): DataLoader of samples.
        max_epochs (int): Number of epochs.
        lr (float): learning rate.
        combinations (list, optional): determines posteriors that are generated.
            examples:
                [[0,1], [3,4]]: p(z_0,z_1) and p(z_3,z_4) are generated
                    initialize network with zdim = 2, pdim = 2
                [[0,1,5,2]]: p(z_0,z_1,z_5,z_2) is generated
                    initialize network with zdim = 1, pdim = 4
        device (str, device): Move batches to this device.
        non_blocking (bool): non_blocking in .to(device) expression.

    Returns:
        list: list of training losses.
    """
    # TODO consider that the user might want other training stats, like number of correct samples for example
    def do_epoch(loader: torch.utils.data.dataloader.DataLoader, train: bool):
        accumulated_loss = 0
        training_context = suppress() if train else torch.no_grad()
        with training_context:
            for batch in loader:
                optimizer.zero_grad()
                if device is not None:
                    batch = {
                        k: v.to(device, non_blocking=non_blocking)
                        for k, v in batch.items()
                    }
                loss = loss_fn(network, batch, combinations=combinations)
                if train:
                    loss.backward()
                    optimizer.step()
                accumulated_loss += loss.detach().cpu().numpy().item()
        return accumulated_loss

    max_epochs = 2 ** 31 - 1 if max_epochs is None else max_epochs
    optimizer = torch.optim.Adam(network.parameters(), lr=lr)

    n_train_batches = len(train_loader) if len(train_loader) != 0 else 1
    n_validation_batches = len(validation_loader) if len(validation_loader) != 0 else 1

    train_losses, validation_losses = [], []
    epoch, fruitless_epoch, min_loss = 0, 0, float("Inf")
    while epoch < max_epochs and fruitless_epoch < early_stopping_patience:
        # print("Epoch:", epoch, end = "\r")
        network.train()
        train_loss = do_epoch(train_loader, True)
        train_losses.append(train_loss / n_train_batches)

        network.eval()
        validation_loss = do_epoch(validation_loader, False)
        print("Validation loss:", validation_loss)
        validation_losses.append(validation_loss / n_validation_batches)

        epoch += 1
        if epoch == 0 or min_loss > validation_loss:
            fruitless_epoch = 0
            min_loss = validation_loss
            best_state_dict = deepcopy(network.state_dict())
        else:
            fruitless_epoch += 1

    print("Total epochs:", epoch)
    # print("Validation losses:", validation_losses)
    return train_losses, validation_losses, best_state_dict


def trainloop(
    net,
    dataset,
    combinations=None,
    batch_size=32,
    nworkers=4,
    max_epochs=50,
    early_stopping_patience=1,
    device="cpu",
    lr_schedule=[1e-3, 1e-4, 1e-5],
    percent_validation=0.1,
):
    print("Start training")
    percent_train = 1.0 - percent_validation
    ntrain, nvalid = split_length_by_percentage(
        len(dataset), (percent_train, percent_validation)
    )
    dataset_train, dataset_valid = torch.utils.data.random_split(
        dataset, [ntrain, nvalid]
    )
    train_loader = torch.utils.data.DataLoader(
        dataset_train,
        batch_size=batch_size,
        num_workers=nworkers,
        pin_memory=True,
        drop_last=True,
    )
    valid_loader = torch.utils.data.DataLoader(
        dataset_valid,
        batch_size=batch_size,
        num_workers=nworkers,
        pin_memory=True,
        drop_last=True,
    )

    # Train!
    train_loss, valid_loss = [], []
    for i, lr in enumerate(lr_schedule):
        print(f"LR iteration {i}")
        tl, vl, sd = train(
            net,
            train_loader,
            valid_loader,
            early_stopping_patience=early_stopping_patience,
            lr=lr,
            max_epochs=max_epochs,
            device=device,
            combinations=combinations,
        )
        vl_minimum = min(vl)
        vl_min_idx = vl.index(vl_minimum)
        train_loss.append(tl[: vl_min_idx + 1])
        valid_loss.append(vl[: vl_min_idx + 1])
        net.load_state_dict(sd)


def get_statistics(
    points: Union["swyft.estimation.Points", Sequence[Dict[str, Array]]],
    combinations: Combinations = None,
    n_samples: int = 300,
):
    """Calculate the mean and std of both x and z.

    Args:
        points: list of dictionaries with keys 'x' and 'z'.
        combinations
        n_samples: size of sample

    Returns:
        x_mean, x_std, z_mean, z_std
    """
    irand = np.random.choice(len(points), n_samples)
    x = [points[i]["x"] for i in irand]
    z = [points[i]["z"] for i in irand]
    x_mean = sum(x) / len(x)
    z_mean = sum(z) / len(z)
    x_var = sum([(x[i] - x_mean) ** 2 for i in range(len(x))]) / len(x)
    z_var = sum([(z[i] - z_mean) ** 2 for i in range(len(z))]) / len(z)

    z_mean = combine_z(z_mean, combinations)
    z_var = combine_z(z_var, combinations)

    return x_mean, x_var ** 0.5, z_mean, z_var ** 0.5


if __name__ == "__main__":
    pass



import pytest

import numpy as np

from swyft.estimation import RatioEstimator
from swyft.eval import eval_net, get_ratios

from .estimation_test import sim_repeat_noise, setup_points, Head


class TestEvaluation:
    @staticmethod
    def setup_re_x0_z0_points():
        _, points = setup_points()
        head = Head(*Head.featurize(points.xshape))
        re = RatioEstimator(points, head=head)

        z0 = np.random.rand(points.zdim)
        x0 = sim_repeat_noise(z0, points.xshape[0])
        z0 = z0[None, :, None]
        x0 = x0
        return re, x0, z0, points

    @pytest.mark.parametrize("training", (True, False))
    def test_eval_net_network_state(self, training):
        re, x0, z0, _ = TestEvaluation.setup_re_x0_z0_points()

        if training:
            re.net.train()
        else:
            re.net.eval()

        _ = eval_net(x0, re.net, z0, batch_size=1)

        assert re.net.training == training

    @pytest.mark.parametrize("training", (True, False))
    def test_get_ratios_network_state(self, training):
        re, x0, _, points = TestEvaluation.setup_re_x0_z0_points()

        if training:
            re.net.train()
        else:
            re.net.eval()

        _ = get_ratios(x0, re.net, points)

        assert re.net.training == training



# pylint: disable=no-member, not-callable
from warnings import warn
from pathlib import Path

import numpy as np
import torch
import scipy


from .types import (
    Optional,
    Device,
    Tensor,
    Array,
    List,
    Sequence,
    Combinations,
    PathType,
)


def get_2d_combinations(indices: List[int]):
    """Given a list of indices, calculate the lower triangular part of the cartesian product.
    Appropriate for retrieving all 2d combinations of indices (up to permutation).

    Args:
        indices: a list of indices
    """
    output = []
    for i in range(len(indices)):
        for j in range(i + 1, len(indices)):
            output.append([indices[i], indices[j]])
    return output


def combine_z(z: Tensor, combinations: Optional[List]) -> Tensor:
    """Generate parameter combinations in last dimension using fancy indexing.

    Args:
        z: parameters of shape [..., Z]
        combinations: list of parameter combinations.

    Returns:
        output = z[..., combinations]
    """
    return z[..., combinations]


def set_device(gpu: bool = False) -> torch.device:
    """Select device, defaults to cpu."""
    if gpu and torch.cuda.is_available():
        device = torch.device("cuda")
        torch.set_default_tensor_type("torch.cuda.FloatTensor")
    elif gpu and not torch.cuda.is_available():
        warn("Although the gpu flag was true, the gpu is not avaliable.")
        device = torch.device("cpu")
        torch.set_default_tensor_type("torch.FloatTensor")
    else:
        device = torch.device("cpu")
        torch.set_default_tensor_type("torch.FloatTensor")
    return device


def get_x(list_xz):
    """Extract x from batch of samples."""
    return [xz["x"] for xz in list_xz]


def get_z(list_xz):
    """Extract z from batch of samples."""
    return [xz["z"] for xz in list_xz]


def get_device_if_not_none(device: Optional[Device], tensor: Tensor) -> Device:
    """Returns device if not None, else returns tensor.device."""
    return tensor.device if device is None else device


np_bool_types = [np.bool]
np_int_types = [np.int8, np.int16, np.int32, np.int64]
np_float_types = [np.float32, np.float64]
torch_bool_types = [torch.bool]
torch_int_types = [torch.int8, torch.int16, torch.int32, torch.int64]
torch_float_types = [torch.float32, torch.float64]


def array_to_tensor(
    array: Array, dtype: Optional[torch.dtype] = None, device: Optional[Device] = None
) -> Tensor:
    """Converts np.ndarray and torch.Tensor to torch.Tensor with dtype and on device.
    When dtype is None, unsafe casts all float-type arrays to torch.float32 and all int-type arrays to torch.int64
    """
    input_dtype = array.dtype
    if isinstance(input_dtype, np.dtype):
        if dtype is None:
            if input_dtype in np_float_types:
                dtype = torch.float32
            elif input_dtype in np_int_types:
                dtype = torch.int64
            elif input_dtype in np_bool_types:
                dtype = torch.bool
            else:
                raise TypeError(
                    f"{input_dtype} was not a supported numpy int, float, or bool."
                )
        return torch.from_numpy(array).to(dtype=dtype, device=device)
    elif isinstance(input_dtype, torch.dtype):
        if dtype is None:
            if input_dtype in torch_float_types:
                dtype = torch.float32
            elif input_dtype in torch_int_types:
                dtype = torch.int64
            elif input_dtype in torch_bool_types:
                dtype = torch.bool
            else:
                raise TypeError(
                    f"{input_dtype} was not a supported torch int, float, or bool."
                )
        return array.to(dtype=dtype, device=device)
    else:
        raise TypeError(
            f"{input_dtype} was not recognized as a supported numpy.dtype or torch.dtype."
        )


def tobytes(x: Array):
    if isinstance(x, np.ndarray):
        return x.tobytes()
    elif isinstance(x, Tensor):
        return x.numpy().tobytes()
    else:
        raise TypeError(f"{type(x)} does not support tobytes.")


def depth(seq: Sequence):
    if seq and isinstance(seq, Sequence):
        return 1 + max(depth(item) for item in seq)
    else:
        return 0


def process_combinations(comb: Combinations):
    d = depth(comb)
    if d == 0:
        return [[comb]]
    elif d == 1:
        return [[i] for i in comb]
    elif d == 2:
        return comb
    else:
        raise ValueError(f"{comb} is not understood to be of type Combinations.")


def is_empty(directory: PathType):
    directory = Path(directory)
    if next(directory.iterdir(), None) is None:
        return True
    else:
        return False


def get_stats(z, p):
    # Returns central credible intervals
    zmax = z[p.argmax()]
    c = scipy.integrate.cumtrapz(p, z, initial=0)
    res = np.interp([0.025, 0.16, 0.5, 0.84, 0.975], c, z)
    xmedian = res[2]
    xerr68 = [res[1], res[3]]
    xerr95 = [res[0], res[4]]
    return {
        "mode": zmax,
        "median": xmedian,
        "cred68": xerr68,
        "cred95": xerr95,
        "err68": (xerr68[1] - xerr68[0]) / 2,
        "err95": (xerr95[1] - xerr95[0]) / 2,
    }


def cred1d(re, x0: Array):
    """Calculate credible regions.

    Args:
        re (RatioEstimator): ratio estimators
        x0: true observation
    """
    zdim = re.zdim
    for i in range(zdim):
        z, p = re.posterior(x0, i)
        res = get_stats(z, p)
        print("z%i = %.5f +- %.5f" % (i, res["median"], res["err68"]))


if __name__ == "__main__":
    pass



# pylint: disable=no-member, not-callable
from warnings import warn
from copy import deepcopy
from functools import wraps
from pathlib import Path
import pickle

import numpy as np
from scipy.integrate import trapz

import torch
import torch.nn as nn

from .train import trainloop
from .network import Network
from .eval import get_ratios
from .types import (
    Sequence,
    Tuple,
    Device,
    Dataset,
    Combinations,
    Callable,
    Array,
    Union,
    PathType,
    Dict,
    Optional,
)
from .utils import array_to_tensor, tobytes, process_combinations


class RatioEstimator:

    _save_attrs = ["net_state_dict", "ratio_cache", "combinations"]

    def __init__(
        self,
        points: Dataset,
        combinations: Optional[Combinations] = None,
        head: Optional[nn.Module] = None,
        previous_ratio_estimator=None,
        device: Device = "cpu",
        statistics=None,
    ):
        """RatioEstimator takes simulated points from the iP3 sample cache and handles training and posterior calculation.

        Args:
            points: points dataset from the iP3 sample cache
            combinations: which combinations of z parameters to learn
            head: initialized module which processes observations, head(x0) = y
            previous_ratio_estimator: ratio estimator from another round. if given, reuse head.
            device: default is cpu
            statistics: x_mean, x_std, z_mean, z_std
        """
        self.points = points
        self._combinations = combinations
        self.head = head if head is None else head.to(device)
        self.prev_re = previous_ratio_estimator
        self.device = device

        self.net = self._init_net(statistics, self.prev_re)
        self.ratio_cache: Dict[bytes : np.ndarray] = {}

    @property
    def zdim(self):
        return self.points.zdim

    @property
    def xshape(self):
        return self.points.xshape

    @property
    def combinations(self) -> Combinations:
        """Marginals that the ratio estimator learns."""
        if self._combinations is None:
            return [[i] for i in range(self.zdim)]
        else:
            return process_combinations(self._combinations)

    def _init_net(self, statistics: Tuple, previous_ratio_estimator):
        """Options for custom network initialization.

        Args:
            statistics: x_mean, x_std, z_mean, z_std
            previous_ratio_estimator: ratio estimator from another round. if given, reuse head.
        """
        # TODO change this to state_dict and deal with the self.head is None case more gracefully.
        if previous_ratio_estimator is not None:
            if self.head is not None:
                warn("using previous ratio estimator's head rather than yours.")
            self.head = deepcopy(self.prev_re.net.head)
        # TODO this is an antipattern address it in network by removing pnum and pdim
        pnum = len(self.combinations)
        pdim = len(self.combinations[0])

        # TODO network should be able to handle shape or dim. right now we are forcing dim, that is bad.
        if self.head is None:
            # yshape = self.xshape
            yshape = self.xshape[0]
        else:
            input_x = array_to_tensor(torch.empty(1, *self.xshape, device=self.device))
            # yshape = self.head(input_x).shape[1:]
            yshape = self.head(input_x).shape[1]
        print("yshape (shape of features between head and legs):", yshape)
        return Network(
            ydim=yshape, pnum=pnum, pdim=pdim, head=self.head, datanorms=statistics
        ).to(self.device)

    def train(
        self,
        max_epochs: int = 100,
        batch_size: int = 8,
        lr_schedule: Sequence[float] = [1e-3, 1e-4, 1e-5],
        early_stopping_patience: int = 1,
        nworkers: int = 0,
        percent_validation=0.1,
    ) -> None:
        """Train higher-dimensional marginal posteriors.

        Args:
            max_epochs: maximum number of training epochs
            batch_size: minibatch size
            lr_schedule: list of learning rates
            early_stopping_patience: early stopping patience
            nworkers: number of Dataloader workers (0 for no dataloader parallelization)
            percent_validation: percentage to allocate to validation set
        """
        trainloop(
            self.net,
            self.points,
            combinations=self.combinations,
            device=self.device,
            max_epochs=max_epochs,
            batch_size=batch_size,
            lr_schedule=lr_schedule,
            early_stopping_patience=early_stopping_patience,
            nworkers=nworkers,
            percent_validation=percent_validation,
        )
        return None

    def _eval_ratios(self, x0: Array, max_n_points: int):
        """Evaluate ratios if not already evaluated.

        Args:
            x0 (array): real observation to calculate posterior
            max_n_points (int): number of points to calculate ratios
        """
        binary_x0 = tobytes(x0)
        if binary_x0 in self.ratio_cache.keys():
            pass
        else:
            z, ratios = get_ratios(
                x0,
                self.net,
                self.points,
                device=self.device,
                combinations=self.combinations,
                max_n_points=max_n_points,
            )
            self.ratio_cache[binary_x0] = {"z": z, "ratios": ratios}
        return None

    def posterior(
        self,
        x0: Array,
        combination_indices: Union[int, Sequence[int]],
        max_n_points: int = 1000,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Retrieve estimated marginal posterior.

        Args:
            x0: real observation to calculate posterior
            combination_indices: z indices in self.combinations
            max_n_points: number of points to calculate ratios on

        Returns:
            parameter array, posterior array
        """
        self._eval_ratios(x0, max_n_points=max_n_points)

        if isinstance(combination_indices, int):
            combination_indices = [combination_indices]

        j = self.combinations.index(combination_indices)
        z = self.ratio_cache[tobytes(x0)]["z"][:, j]
        ratios = self.ratio_cache[tobytes(x0)]["ratios"][:, j]

        # 1-dim case
        if len(combination_indices) == 1:
            z = z[:, 0]
            isorted = np.argsort(z)
            z, ratios = z[isorted], ratios[isorted]
            exp_r = np.exp(ratios)
            I = trapz(exp_r, z)
            p = exp_r / I
        else:
            p = np.exp(ratios)
        return z, p

    @property
    def net_state_dict(self):
        return self.net.state_dict()

    @property
    def _save_dict(self):
        return {attr: getattr(self, attr) for attr in RatioEstimator._save_attrs}

    def save(self, path: PathType):
        complete = {}
        complete.update(self.points._save_dict)
        complete.update(self._save_dict)

        path = Path(path)
        with path.open("wb") as f:
            pickle.dump(complete, f)
        return None

    @classmethod
    def load(
        cls,
        cache: "swyft.cache.Cache",
        path: PathType,
        head: nn.Module = None,
        noisehook: Callable = None,
        device: Device = None,
    ):
        """Load pickled ratio estimator.

        Args:
            cache: cache object on which the estimator was trained
            path: path to pickled ratio estimator
            head: head network utilized in pickled ratio estimator
            noisehook: maps from (x, z) to x with noise
            device
        """
        path = Path(path)
        with path.open("rb") as f:
            complete = pickle.load(f)

        point_attrs = {attr: complete[attr] for attr in Points._save_attrs}
        points = Points._load(cache, point_attrs, noisehook)

        re_attrs = {attr: complete[attr] for attr in RatioEstimator._save_attrs}
        instance = cls(points, re_attrs["combinations"], head, device)
        instance.net.load_state_dict(
            re_attrs["net_state_dict"],
        )
        instance.ratio_cache = re_attrs["ratio_cache"]
        return instance


class Points(torch.utils.data.Dataset):
    """Points references (observation, parameter) pairs drawn from an inhomogenous Poisson Point Proccess (iP3) Cache.
    Points implements this via a list of indices corresponding to data contained in a cache which is provided at initialization.
    """

    _save_attrs = ["intensity", "indices"]

    def __init__(self, cache: "swyft.cache.Cache", intensity, noisehook=None):
        """Create a points dataset

        Args:
            cache (Cache): iP3 cache for zarr storage
            intensity (Intensity): inhomogenous Poisson Point Proccess intensity function on parameters
            noisehook (function): (optional) maps from (x, z) to x with noise
        """
        super().__init__()
        if cache.requires_sim():
            raise RuntimeError(
                "The cache has parameters without a corresponding observation. Try running the simulator."
            )

        self.cache = cache
        self.intensity = intensity
        self.noisehook = noisehook
        self._indices = None

    def __len__(self):
        assert len(self.indices) <= len(
            self.cache
        ), f"You wanted {len(self.indices)} indices but there are only {len(self.cache)} parameter samples in the cache."
        return len(self.indices)

    def __getitem__(self, idx):
        i = self.indices[idx]
        x = self.cache.x[i]
        z = self.cache.z[i]

        if self.noisehook is not None:
            x = self.noisehook(x, z)

        x = torch.from_numpy(x).float()
        z = torch.from_numpy(z).float()
        return {
            "x": x,
            "z": z,
        }

    @property
    def zdim(self):
        assert (
            self.intensity.zdim == self.cache.zdim
        ), "The cache and intensity functions did not agree on the zdim."
        return self.intensity.zdim

    @property
    def xshape(self):
        return self.cache.xshape

    @property
    def indices(self):
        if self._indices is None:
            self._indices = self.cache.sample(self.intensity)
            self._check_fidelity_to_cache(self._indices)
        return self._indices

    def _check_fidelity_to_cache(self, indices):
        first = indices[0]
        assert self.zdim == len(
            self.cache.z[first]
        ), "Item in cache did not agree with zdim."
        assert (
            self.xshape == self.cache.x[first].shape
        ), "Item in cache did not agree with xshape."
        return None

    @property
    def _save_dict(self):
        return {attr: getattr(self, attr) for attr in Points._save_attrs}

    def save(self, path: PathType):
        """Saves indices and intensity in a pickle."""
        path = Path(path)
        with path.open("wb") as f:
            pickle.dump(self._save_dict, f)
        return None

    @classmethod
    def _load(cls, cache, attrs: dict, noisehook):
        instance = cls(cache, attrs["intensity"], noisehook=noisehook)
        instance._indices = attrs["indices"]
        instance._check_fidelity_to_cache(attrs["indices"])
        return instance

    @classmethod
    def load(cls, cache: "swyft.cache.Cache", path: PathType, noisehook=None):
        """Loads saved indices and intensity from a pickle. User provides cache and noisehook."""
        path = Path(path)
        with path.open("rb") as f:
            attrs = pickle.load(f)
        return cls._load(cache, attrs, noisehook)



# pylint: disable=no-member, not-callable
from abc import ABC, abstractmethod
from pathlib import Path

import numpy as np
import zarr
import numcodecs
from tqdm import tqdm

from .utils import is_empty
from .types import Shape, Union, Sequence, PathType, Callable, Array


class Cache(ABC):
    """Abstract base class for various caches."""

    _filesystem = (
        "metadata",
        "metadata/intensity",
        "metadata/requires_simulation",
        "samples",
        "samples/x",
        "samples/z",
    )

    @abstractmethod
    def __init__(
        self,
        zdim: int,
        xshape: Shape,
        store: Union[zarr.MemoryStore, zarr.DirectoryStore],
    ):
        """Initialize Cache content dimensions.

        Args:
            zdim: Number of z dimensions
            xshape: Shape of x array
            store: zarr storage.
        """
        self._zdim = None
        self._xshape = None
        self.store = store
        self.root = zarr.group(store=self.store)

        if all(key in self.root.keys() for key in ["samples", "metadata"]):
            print("Loading existing cache.")
            self._update()
        elif len(self.root.keys()) == 0:
            print("Creating new cache.")
            self.x = self.root.zeros(
                self._filesystem[4],  # samples/x
                shape=(0, *xshape),
                chunks=(1, *xshape),
                dtype="f4",
            )
            self.z = self.root.zeros(
                self._filesystem[5],  # samples/z
                shape=(0, zdim),
                chunks=(10000, zdim),
                dtype="f4",
            )
            self.m = self.root.zeros(
                self._filesystem[2],  # metadata/requires_simulation
                shape=(0, 1),
                chunks=(10000, 1),
                dtype="bool",
            )
            self.u = self.root.create(
                self._filesystem[1],  # metadata/intensity
                shape=(0,),
                dtype=object,
                object_codec=numcodecs.Pickle(),
            )
        else:
            raise KeyError(
                "The zarr storage is corrupted. It should either be empty or only have the keys ['samples', 'metadata']."
            )

        assert (
            zdim == self.zdim
        ), f"Your given zdim, {zdim}, was not equal to the one defined in zarr {self.zdim}."
        assert (
            xshape == self.xshape
        ), f"Your given xshape, {xshape}, was not equal to the one defined in zarr {self.xshape}."

    @staticmethod
    def _extract_xshape_from_zarr_group(group):
        return group["samples/x"].shape[1:]

    @staticmethod
    def _extract_zdim_from_zarr_group(group):
        return group["samples/z"].shape[1]

    @property
    def xshape(self) -> Shape:
        """Shape of observations."""
        if self._xshape is None:
            self._xshape = self._extract_xshape_from_zarr_group(self.root)
        return self._xshape

    @property
    def zdim(self) -> int:
        """Dimension of parameters."""
        if self._zdim is None:
            self._zdim = self._extract_zdim_from_zarr_group(self.root)
        return self._zdim

    def _update(self):
        # This could be removed with a property for each attribute which only loads from disk if something has changed. TODO
        self.x = self.root["samples/x"]
        self.z = self.root["samples/z"]
        self.m = self.root["metadata/requires_simulation"]
        self.u = self.root["metadata/intensity"]

    def __len__(self):
        """Returns number of samples in the cache."""
        self._update()
        return len(self.z)

    def __getitem__(self, i):
        self._update()
        return self.x[i], self.z[i]

    def _append_z(self, z):
        """Append z to cache content and new slots for x."""
        self._update()

        # Add simulation slots
        xshape = list(self.x.shape)
        xshape[0] += len(z)
        self.x.resize(*xshape)

        # Add z samples
        self.z.append(z)

        # Register as missing
        m = np.ones((len(z), 1), dtype="bool")
        self.m.append(m)

    def intensity(self, z: Array) -> np.ndarray:
        """Evaluate the cache's intensity function.

        Args:
            z: list of parameter values.
        """
        self._update()

        if len(self.u) == 0:
            return np.zeros(len(z))
        else:
            return np.array([self.u[i](z) for i in range(len(self.u))]).max(axis=0)

    def grow(self, intensity: "swyft.intensity.Intensity"):
        """Given an intensity function, add parameter samples to the cache.

        Args:
            intensity: target parameter intensity function
        """
        # Proposed new samples z from p
        z_prop = intensity.sample()

        # Rejection sampling from proposal list
        accepted = []
        ds_intensities = self.intensity(z_prop)
        target_intensities = intensity(z_prop)
        for _, Ids, It in zip(z_prop, ds_intensities, target_intensities):
            rej_prob = np.minimum(1, Ids / It)
            w = np.random.rand()
            accepted.append(rej_prob < w)
        z_accepted = z_prop[accepted, :]

        # Add new entries to cache
        if len(z_accepted) > 0:
            self._append_z(z_accepted)
            print("Adding %i new samples. Run simulator!" % len(z_accepted))
        else:
            print("No new simulator runs required.")

        # save intensity function. We collect them all to find their maximum.
        self.u.resize(len(self.u) + 1)
        self.u[-1] = intensity

    def sample(self, intensity: "swyft.intensity.Intensity"):
        """Sample from Cache.

        Args:
            intensity: target parameter intensity function
        """
        self._update()

        self.grow(intensity)

        accepted = []
        zlist = self.z[:]
        I_ds = self.intensity(zlist)
        I_target = intensity(zlist)
        for i, z in enumerate(zlist):
            accept_prob = I_target[i] / I_ds[i]
            assert accept_prob <= 1.0, (
                f"{accept_prob} > 1, but we expected the ratio of target intensity function to the cache <= 1. "
                "There may not be enough samples in the cache "
                "or a constrained intensity function was not accounted for."
            )
            w = np.random.rand()
            if accept_prob > w:
                accepted.append(i)
        return accepted

    def _require_sim_idx(self):
        indices = []
        m = self.m[:]
        for i in range(len(self.z)):
            if m[i]:
                indices.append(i)
        return indices

    def requires_sim(self) -> bool:
        """Check whether there are parameters which require a matching simulation."""
        self._update()

        return len(self._require_sim_idx()) > 0

    def _add_sim(self, i, x):
        self.x[i] = x
        self.m[i] = False

    def simulate(self, simulator: Callable):
        """Run simulator sequentially on parameter cache with missing corresponding simulations.

        Args:
            simulator: simulates an observation given a parameter input
        """
        self._update()

        idx = self._require_sim_idx()
        if len(idx) == 0:
            print("No simulations required.")
            return
        for i in tqdm(idx, desc="Simulate"):
            z = self.z[i]
            x = simulator(z)
            self._add_sim(i, x)


class DirectoryCache(Cache):
    def __init__(self, zdim: int, xshape: Shape, path: PathType):
        """Instantiate an iP3 cache stored in a directory.

        Args:
            zdim: Number of z dimensions
            xshape: Shape of x array
            path: path to storage directory
        """
        self.store = zarr.DirectoryStore(path)
        super().__init__(zdim=zdim, xshape=xshape, store=self.store)

    @classmethod
    def load(cls, path: PathType):
        """Load existing DirectoryStore."""
        store = zarr.DirectoryStore(path)
        group = zarr.group(store=store)
        xshape = cls._extract_xshape_from_zarr_group(group)
        zdim = cls._extract_zdim_from_zarr_group(group)
        return DirectoryCache(zdim=zdim, xshape=xshape, path=path)


class MemoryCache(Cache):
    def __init__(self, zdim: int, xshape: Shape, store=None):
        """Instantiate an iP3 cache stored in the memory.

        Args:
            zdim: Number of z dimensions
            xshape: Shape of x array
            store (zarr.MemoryStore, zarr.DirectoryStore): optional, used in loading.
        """
        if store is None:
            self.store = zarr.MemoryStore()
        else:
            self.store = store
        super().__init__(zdim=zdim, xshape=xshape, store=self.store)

    def save(self, path: PathType) -> None:
        """Save the current state of the MemoryCache to a directory."""
        path = Path(path)
        if path.exists() and not path.is_dir():
            raise NotADirectoryError(f"{path} should be a directory")
        elif path.exists() and not is_empty(path):
            raise FileExistsError(f"{path} is not empty")
        else:
            path.mkdir(parents=True, exist_ok=True)
            store = zarr.DirectoryStore(path)
            zarr.convenience.copy_store(source=self.store, dest=store)
            return None

    @classmethod
    def load(cls, path: PathType):
        """Load existing DirectoryStore state into a MemoryCache object."""
        memory_store = zarr.MemoryStore()
        directory_store = zarr.DirectoryStore(path)
        zarr.convenience.copy_store(source=directory_store, dest=memory_store)

        group = zarr.group(store=memory_store)
        xshape = cls._extract_xshape_from_zarr_group(group)
        zdim = cls._extract_zdim_from_zarr_group(group)
        return MemoryCache(zdim=zdim, xshape=xshape, store=memory_store)


if __name__ == "__main__":
    pass



from itertools import product

import pytest
import glob
import tempfile

import torch
from swyft.cache import Cache, MemoryCache, DirectoryCache


class TestCacheIO:
    zdims = [0, 1, 5]
    xshapes = [(0,), (1,), (4, 6, 2, 1)]

    @pytest.mark.parametrize("zdim, xshape", product(zdims, xshapes))
    def test_memory_cache_save(self, zdim, xshape):
        cache = MemoryCache(zdim, xshape)
        with tempfile.TemporaryDirectory() as td:
            cache.save(td)
            items = glob.glob(td + "/**", recursive=True)

        without_prefix = sorted([item[len(td) + 1 :] for item in items])
        without_blanks = sorted([item for item in without_prefix if item])
        assert all(
            [item == truth for item, truth in zip(without_blanks, Cache._filesystem)]
        )

    @pytest.mark.parametrize("zdim, xshape", product(zdims, xshapes))
    def test_memory_cache_load(self, zdim, xshape):
        cache = MemoryCache(zdim, xshape)
        cache_states = (cache.zdim, cache.xshape, len(cache))

        with tempfile.TemporaryDirectory() as td:
            cache.save(td)
            loaded = MemoryCache.load(td)
            loaded_stats = (loaded.zdim, loaded.xshape, len(loaded))
            assert cache_states == loaded_stats

    @pytest.mark.parametrize("zdim, xshape", product(zdims, xshapes))
    def test_directory_cache_load(self, zdim, xshape):
        cache = MemoryCache(zdim, xshape)
        cache_states = (cache.zdim, cache.xshape, len(cache))

        with tempfile.TemporaryDirectory() as td:
            cache.save(td)
            loaded = DirectoryCache.load(td)
            loaded_stats = (loaded.zdim, loaded.xshape, len(loaded))
            assert cache_states == loaded_stats


if __name__ == "__main__":
    pass



# pylint: disable=no-member
from itertools import product

import pytest

import torch
import numpy as np
from swyft.utils import (
    array_to_tensor,
    np_bool_types,
    np_int_types,
    np_float_types,
    torch_bool_types,
    torch_int_types,
    torch_float_types,
)


class TestArrayToTensor:
    shapes = [
        (0,),
        (0, 0),
        (1, 0),
        (0, 1),
        (10, 10),
        (15, 3, 2, 1, 6),
    ]

    @pytest.mark.parametrize(
        "shape, dtype",
        product(shapes, np_bool_types + np_int_types + np_float_types + [None]),
    )
    def test_shape(self, shape, dtype):
        array = np.empty(shape, dtype)
        tensor = array_to_tensor(array)
        assert tensor.shape == array.shape

    @pytest.mark.parametrize(
        "shape, dtypes",
        product(
            shapes,
            zip(
                np_int_types + np_int_types,
                torch_int_types + [None] * len(np_int_types),
            ),
        ),
    )
    def test_np_to_tensor_int(self, shape, dtypes):
        np_dtype, torch_dtype = dtypes
        array = np.empty(shape, np_dtype)
        tensor = array_to_tensor(array, dtype=torch_dtype)
        if torch_dtype is None:
            assert tensor.dtype == torch.int64
        else:
            assert tensor.dtype == torch_dtype

    @pytest.mark.parametrize("shape, dtype", product(shapes, np_bool_types))
    def test_np_to_tensor_bool(self, shape, dtype):
        array = np.empty(shape, dtype)
        tensor = array_to_tensor(array)
        assert tensor.dtype == torch.bool

    @pytest.mark.parametrize(
        "shape, np_dtype, torch_dtype",
        product(shapes, np_float_types, torch_float_types + [None]),
    )
    def test_np_to_tensor_float(self, shape, np_dtype, torch_dtype):
        array = np.empty(shape, np_dtype)
        tensor = array_to_tensor(array, dtype=torch_dtype)
        if torch_dtype is None:
            assert tensor.dtype == torch.float32
        else:
            assert tensor.dtype == torch_dtype


if __name__ == "__main__":
    pass



import numpy as np

import pylab as plt
from scipy.interpolate import griddata

from .types import Array, Tuple, Sequence


def get_contour_levels(x, cred_level=[0.68268, 0.95450, 0.99730]):
    x = np.sort(x)[::-1]  # Sort backwards
    total_mass = x.sum()
    enclosed_mass = np.cumsum(x)
    idx = [np.argmax(enclosed_mass >= total_mass * f) for f in cred_level]
    levels = np.array(x[idx])
    return levels


def cont2d(
    ax,
    re: "swyft.estimation.RatioEstimator",
    x0: Array,
    z0: Array,
    i: int,
    j: int,
    cmap: str = "gray_r",
    max_n_points: int = 1000,
):
    """Create a 2d contour plot.

    Args:
        ax (matplotlib.axes.Axes): matplotlib axes
        re: train ratio estimator
        x0: true observation
        z0: true parameters
        i: combination index
        j: combination index
        cmap: color map
        max_n_points: number of points to train on
    """
    z, p = re.posterior(x0, [i, j], max_n_points=max_n_points)
    levels = get_contour_levels(p)

    if z0 is not None:
        ax.axvline(z0[i], color="r", ls=":")
        ax.axhline(z0[j], color="r", ls=":")

    N = 100 * 1j
    extent = [z[:, 0].min(), z[:, 0].max(), z[:, 1].min(), z[:, 1].max()]
    xs, ys = np.mgrid[
        z[:, 0].min() : z[:, 0].max() : N, z[:, 1].min() : z[:, 1].max() : N
    ]
    resampled = griddata(z, p, (xs, ys))
    ax.imshow(resampled.T, extent=extent, origin="lower", cmap=cmap, aspect="auto")
    ax.tricontour(z[:, 0], z[:, 1], -p, levels=-levels, colors="k", linestyles=["-"])


def hist1d(ax, re, x0, z0, i, max_n_points=1000):
    if z0 is not None:
        ax.axvline(z0[i], color="r", ls=":")
    z, p = re.posterior(x0, i, max_n_points=max_n_points)
    ax.plot(z, p, "k")


def plot1d(
    re1d: "swyft.estimation.RatioEstimator",
    x0: Array,
    dims: Tuple[int, int] = (15, 5),
    ncol: int = None,
    params: Sequence[str] = None,
    labels: Sequence[str] = None,
    z0: Array = None,
    cmap: str = "Greys",
    max_n_points: int = 1000,
):
    """Create a one dimensional plot.

    Args:
        re1d:
        x0:
        dims: dimension of output plot
        ncol: =None
        params: =None
        labels: =None,
        z0: =None,
        cmap: color map
        max_n_points: number of points to train on
    """
    # TODO: Rewrite
    if params is None:
        params = range(re1d.zdim)

    if ncol is None:
        ncol = re1d.zdim

    K = len(params)
    nrow = (K - 1) // ncol + 1

    fig, axes = plt.subplots(nrow, ncol, figsize=dims)
    lb = 0.125
    tr = 0.9
    whspace = 0.15
    fig.subplots_adjust(
        left=lb, bottom=lb, right=tr, top=tr, wspace=whspace, hspace=whspace
    )

    if labels is None:
        labels = ["z%i" % params[i] for i in range(K)]
    for k in range(K):
        if nrow == 1:
            ax = axes[k]
        else:
            i, j = k % ncol, k // ncol
            ax = axes[j, i]
        hist1d(ax, re1d, x0, z0, params[k], max_n_points=max_n_points)
        ax.set_xlabel(labels[k])


def corner(
    re1d: "swyft.estimation.RatioEstimator",
    re2d: "swyft.estimation.RatioEstimator",
    x0: Array,
    dim: int = 10,
    params: Sequence[str] = None,
    labels: Sequence[str] = None,
    z0: Array = None,
    cmap: str = "Greys",
    max_n_points: int = 1000,
) -> None:
    # TODO: Rewrite
    if params is None:
        params = range(re1d.zdim)

    K = len(params)
    fig, axes = plt.subplots(K, K, figsize=(dim, dim))
    lb = 0.125
    tr = 0.9
    whspace = 0.1
    fig.subplots_adjust(
        left=lb, bottom=lb, right=tr, top=tr, wspace=whspace, hspace=whspace
    )

    if labels is None:
        labels = ["z%i" % params[i] for i in range(K)]
    for i in range(K):
        for j in range(K):
            ax = axes[i, j]
            # Switch off upper left triangle
            if i < j:
                ax.set_yticklabels([])
                ax.set_xticklabels([])
                ax.set_xticks([])
                ax.set_yticks([])
                ax.set_frame_on(False)
                continue

            # Formatting labels
            if j > 0 or i == 0:
                ax.set_yticklabels([])
                # ax.set_yticks([])
            if i < K - 1:
                ax.set_xticklabels([])
                # ax.set_xticks([])
            if i == K - 1:
                ax.set_xlabel(labels[j])
            if j == 0 and i > 0:
                ax.set_ylabel(labels[i])

            # Set limits
            # ax.set_xlim(x_lims[j])
            # if i != j:
            #    ax.set_ylim(y_lims[i])

            # 2-dim plots
            if j < i:
                cont2d(
                    ax,
                    re2d,
                    x0,
                    z0,
                    params[j],
                    params[i],
                    cmap=cmap,
                    max_n_points=max_n_points,
                )

            if j == i:
                hist1d(ax, re1d, x0, z0, params[i], max_n_points=max_n_points)


if __name__ == "__main__":
    pass
